#include "basistools.h"
#include "code.h"
/*
typedef struct {
	int s, t, m;
	vector  *bitbase;
	boole   *basis;
	int     dimSpace;
	size_t  spaceSize;
	vector *dirac;
} basis_t;
*/


void freeBasis( basis_t b )
{
	freecode( b.mat );
	free( b.dirac ) ;
	if ( b.table ) free( b.table ) ;
}

vector decompose( boole vec, code G )
{
vector res = 0;
int i, p;
galois x;
for( i = 0; i < G.nbl && G.piv[i]>=0 ; i++ ){
   p = G.piv[i];
   if ( vec[ p ] )  {
        for( x = 0; x < G.nbc; x++ )
                vec[x] ^= G.fct[i][x];
   	res ^= ((vector) 1) << i;
   }
}
return res;
}

void rsreduction( boole f, int s, int t, int q)
{
        int u; 
        xform( f, q );
        for( u=0; u < q; u++ ) 
                if (  weight( u ) < s  ||  weight( u ) > t ) f[u] = 0;
        xform( f, q );
}



basis_t monomialBasis( int s, int t, int m )
{
	int a;
	basis_t res;
	boole f;
	res.s = s;
	res.t = t;
	res.m = m;
	res.q =  1 << m;
	res.mat    =  rmcode(s, t, m ) ;
	res.size =  1;
	res.size <<= res.mat.nbl;
	res.table = NULL;
	pivotage( res.mat );
//	pcode( "cod", res.mat );
	res.dirac = calloc( res.q , sizeof(vector) ); 
	for( a = 0; a < res.q; a++ ){
		  f = calloc( res.q, sizeof( uchar ) );
		  f[ a ] = 1;
		  rsreduction( f, s, t, 1<<m );
		  res.dirac[a] = decompose( f, res.mat );
		  assert( degree(f)  < 0 );
		  free( f ); 
	}
        return res;	
}


void saveBase( basis_t b)
{
    FILE *dstFile;
    char fn[128];
    size_t size = ( (size_t) 1 ) << b.mat.nbl;
    sprintf( fn, BASEDIR"base-%d-%d-%d.dat", b.s, b.t, b.m );

    dstFile = fopen(fn , "w");

    if ( ! dstFile ) {
	    perror( fn );
	    puts("in saveBase");
	    exit(1);
    }
    if (fwrite(b.table, 1, size, dstFile) != size)
	printf("problem\n");
    fclose(dstFile);
}

void saveBasename( char* name, basis_t b)
{
    FILE *dstFile;
    char fn[128];
    size_t size = ( (size_t) 1 ) << b.mat.nbl;
    sprintf( fn, BASEDIR"base-%s-%d-%d-%d.dat", name, b.s, b.t, b.m );
    dstFile = fopen(fn , "w");
    if ( ! dstFile ) {
	    perror( fn );
	    puts("in saveBasename");
	    exit(1);
    }
    if (fwrite(b.table, 1, size, dstFile) != size)
	printf("problem\n");
    fclose(dstFile);
}

int  checkBasename( char* name, basis_t * b )
{
    FILE *srcFile;
    char fn[128];
    sprintf( fn, BASEDIR"base-%s-%d-%d-%d.dat", name, b->s, b->t, b->m );
    srcFile = fopen( fn, "r");

    if (!srcFile) {
	perror( fn );
	return 0;
    }
    fclose(srcFile);
    return 1;
}

void loadBasename( char* name, basis_t * b )
{
    FILE *srcFile;
    size_t lus;
    char fn[128];
    size_t size = ( (size_t) 1 ) << b->mat.nbl;
    sprintf( fn, BASEDIR"base-%s-%d-%d-%d.dat", name, b->s, b->t, b->m );
    srcFile = fopen( fn, "r");

    b->table= calloc(size, sizeof(unsigned char));
    if (!  b->table) {
	perror("allocation table");
	exit(1);
    }

    if (!srcFile) {
	perror("loadtable");
	exit(1);
    }
    lus = fread( b->table , 1, size, srcFile);
    if (lus != size) {
	printf("load problem:-(\n");
	printf("lus=%ld spaceSize=%ld\n", lus, size);
	exit(1);
    }
    fclose(srcFile);
}
void loadBase( basis_t * b )
{
    FILE *srcFile;
    size_t lus;
    char fn[128];
    size_t size = ( (size_t) 1 ) << b->mat.nbl;
    sprintf( fn, BASEDIR"base-%d-%d-%d.dat", b->s, b->t, b->m );
    srcFile = fopen( fn, "r");

    b->table= calloc(size, sizeof(unsigned char));
    if (!  b->table) {
	perror("allocation table");
	exit(1);
    }

    if (!srcFile) {
	perror("loadtable");
	exit(1);
    }
    lus = fread( b->table , 1, size, srcFile);
    if (lus != size) {
	printf("load problem:-(\n");
	printf("lus=%ld spaceSize=%ld\n", lus, size);
	exit(1);
    }
    fclose(srcFile);
}

void infoBase( basis_t * b , int verb  )
{
size_t t[256]={0}, max  =  0;
size_t r, size = ( (size_t) 1 ) << b->mat.nbl;
for( r = 0; r < size; r++ ) {
	if ( b->table[r] > max ) 
		max = b->table[r];
	t[ b->table[r] ]++;
	}
b->rank = max + 1;
printf("\n#Classification B(%d,%d,%d) :\n#\t dimen=%d\t rank=%d\n", b->s, b->t, b->m, b->mat.nbl, b->rank );
if ( verb ) 
	for( r = 0; r < 256 ; r++ )
       		if ( t[r] )	
	   		printf("%lu -> %lu\n", r, t[r] );
}





















