#include "boolean.h"
#include <ctype.h>

int ffsize = 1, ffdimen = 0;

void initboole( int m )
{
	ffdimen = m;
	ffsize  = 1 << m;
}

boole getboole( void )
{
return calloc ( ffsize , sizeof( uchar)  );
}

boole getboolecpy( boole f )
{ boole res = calloc ( ffsize , sizeof( uchar)  );
  galois x;
  for( x = 0; x < ffsize; x++ )
	  res[x] = f[x];
  return res;
}

int boolecmp( boole f, boole g )
{ galois x;
  for( x=0; x < ffsize && f[x] == g[x] ; x++ ) ;
  return x == ffsize;
}

int linearity( boole f )
{ int tfr[ ffsize ];
  int x;
  int res = 0;
  for( x = 0; x < ffsize; x++ )
	  tfr[x] = f[x] ? -1 : +1;
  Fourier( tfr, ffsize );
  for( x = 0; x < ffsize; x++ )
     if ( abs( tfr[x] ) >  res ) 
	     res =  abs( tfr[x] );
  return res;
}

int isnearbent( boole f )
{ int tfr[ ffsize ];
  int x, val;
  val = 1 << ( (ffdimen+1) / 2 );
  for( x = 0; x < ffsize; x++ )
	  tfr[x] = f[x] ? -1 : +1;
  Fourier( tfr, ffsize );
  for( x = 0; x < ffsize; x++ )
     if (tfr[x] != 0 && tfr[x] != val && tfr[x] != -val)
            return 0;
  return 1;
}

int isbent( boole f )
{ int tfr[ ffsize ];
  int x, val;
  val = 1 << ( ffdimen / 2 );
  for( x = 0; x < ffsize; x++ )
	  tfr[x] = f[x] ? -1 : +1;
  Fourier( tfr, ffsize );
  for( x = 0; x < ffsize; x++ )
     if ( tfr[x] != val && tfr[x] != -val)
            return 0;
  return 1;
}
void xform( boole f, int n)
// Transformation polynomiale sur place de la Booleenne f.
{
int x , y;
if ( n > 1 ) {
         xform(f,n>>1);
         xform(&(f[n>>1]), n>>1);
         for(x=0, y=(n>>1);  x< (n>>1); x++, y++) 
		 f[y] = f[x]^f[y];
         }
}

int weight( galois x )
{ int res = 0;
while ( x ) {
	x &=(x-1);
	res++;
}
return res;
}

int valuation(boole f)
// Retourne la valuation de la Booleene f.
{
int x,p,d;
boole aux;
d = ffdimen;
aux = getboolecpy( f );

xform(aux, ffsize);

for( x=0; x < ffsize; x++)
   if ( aux[x] == 1 )
        {
        p = weight( x );
        if ( p < d  ) d=p;
        }
free(aux);
return(d);
}

int degree(boole f)
// Retourne le degree de la Booleene f.
{
int x,p,d;
boole aux;
d=-1;
aux = getboolecpy( f );

xform(aux, ffsize);

for( x=0; x < ffsize; x++)
   if ( aux[x] == 1 )
        {
        p = weight( x );
        if ( p>d) d=p;
        }
free(aux);
return(d);
}

uint64_t aglCardinality( int  m )
{ uint64_t res = 1;
  int i;
  for( i = 0; i < m; i++ )
          res *=  ( 1LU << m )  - ( 1LU <<  i );
  return res << m;
}

uint64_t glCardinality( int  m )
{ uint64_t res = 1;
  int i;
  for( i = 0; i < m; i++ )
          res *=  ( 1LU << m )  - ( 1LU <<  i );
  return res;
}

uint64_t aglIndex( int  m , uint64_t index )
{ uint64_t res;
  mpz_t rop;
  mpz_init( rop );
  mpz_set_ui( rop, 1); 
  int i;
  for( i = 0; i < m; i++ )
          mpz_mul_ui( rop, rop,   ( 1LU << m )  - ( 1LU <<  i ) );
  mpz_mul_ui( rop, rop,  1LU << m );
  mpz_divexact_ui ( rop , rop, index );
  res =  mpz_get_ui ( rop );
  return res;
}


int anfdimen( char* str )
{ int r = 'a';
  while ( *str ) {
          if ( isalpha(*str) && *str > r)
                 r = *str;
          if ( *str == '=' ) r = 'a';
          str++;
  }
  return r - 'a' + 1;
}


void projboole( int i, int j, boole f)
{ galois x;
  int wt;
  xform( f, ffsize);
  for( x = 0; x < ffsize; x++ ){
    wt = weight( x );
    if ( wt < i || wt > j ) f[x] = 0;
  }
  xform( f, ffsize);
}




void pTT( FILE*dst, boole f)
// prints the anf of f
{
int x;

fprintf(dst, "\nTT=");

for( x = 0; x < ffsize ; x++)
   	fprintf(dst, " %d", f[x]);
        

}


void xprint( FILE*dst, boole f)
// prints the anf of f
{
int x,y,i,p;
boole aux = getboolecpy ( f );
fprintf(dst, "anf=");
p = 1;
xform( aux, ffsize);
if ( aux[0] ) 
	fprintf( dst, "1+" );
for( x = 1; x < ffsize ; x++)
   if ( aux[x] == 1 ) {
        i=0;
        y=x;
        if (p==0) fprintf(dst, "+"); else p=0;
        while (y>0) 
              {
              if ( y & 1 ) fprintf(dst, "%c", i + 'a');
              i++;
              y=y>>1;
              }
        }
free(aux);
}

void panf( FILE*dst, boole f)
{
     fprintf( dst, "\n");
     xprint( dst, f );
}

void panfvalue( FILE*dst, int v, boole f)
{
     fprintf( dst, "\n %d ", v );
     xprint( dst, f );
}
boole strtoboole( char *bfr )
{ char   *ptr;
  galois x = 0;
  int v = 0;
  boole res;
  res = getboole();
  ptr = bfr;
  while ( *ptr && *ptr != '=') ptr++;
  ptr++;
  while ( *ptr ) {
    if ( isdigit( *ptr) ) {
      v  = *ptr - '0';
      x  = 0;
    }
    if ( isalpha( *ptr) ) x ^= 1 << (*ptr - 'a');
    if ( *ptr == '+' ) {
      if ( x == 0 ) res[ x ] = v; else res[ x ] = 1;
      x = 0;
    }
    ptr++;
  }
  if ( x == 0 ) res[ x ] = v; else res[ x ] = 1;
  xform( res, ffsize );
  return res;
}

boole anotherstrtoboole( char *bfr )
{ char   *ptr;
  galois x = 0;
  int v = 0;
  boole res;
  res = getboole();
  ptr = bfr;
  while ( *ptr ) {
    if ( isdigit( *ptr) ) {
      v  = *ptr - '0';
      x  = 0;
    }
    if ( isalpha( *ptr) ) x ^= 1 << (*ptr - 'a');
    if ( *ptr == '+' ) {
      if ( x == 0 ) res[ x ] = v; else res[ x ] = 1;
      x = 0;
    }
    ptr++;
  }
  if ( x == 0 ) res[ x ] = v; else res[ x ] = 1;
  xform( res, ffsize );
  return res;
}

int weightBoole(boole f)
{
    int res = 0;
    galois x;
    for (x = 0; x < ffsize; x++)
	res += f[x];
    return res;
}


boole loadBooleStab(FILE *src, uint64_t  *sz)
//-get the current boolean function
{ boole res = NULL;
  char line[1024];
  while (! feof (src) )
    {
      line[0] = '\0';
      fgets (line, 1024, src);
      switch (*line)
        {
        case '#':
          break;
        case 'a' :
          res =  strtoboole ( line );
	  //return res;
          break;
        case 's' :
          sscanf(line, "stabSize=%ld", sz );
	  return res;
          break;
        default:
	  ;
        }
    }
    return NULL;
}

boole loadBoole(FILE *src )
//-get the current boolean function
{ boole res = NULL;
  char line[1024];
  while (! feof (src) )
    {
      line[0] = '\0';
      fgets (line, 1024, src);
      switch (*line)
        {
        case '#':
          break;
        case 'a' :
          res =  strtoboole ( line );
	  return res;
          break;
        default:;
        }
    }
    return NULL;
}

void Fourier(int *f, unsigned int n )
// Transformation de Fourier sur place.
{
int x,y,z;
if (n>1) {
         Fourier(f,n>>1);
         Fourier(&(f[n>>1]),n>>1);
         for(x=0,y=(n>>1);  x< (n>>1); x++,y++)
                {
                z = f[x];
                f[x] = z + f[y];
                f[y] = z - f[y];
                }
         };
}

int64_t  norme( boole f )
{
long long int sum = 0, tmp;
int  * fct = calloc( ffsize, sizeof(int) );
galois x;
for( x = 0; x < ffsize; x++ )
	fct [ x ] = f[x] ? -1 : + 1;

Fourier( fct, ffsize);
for( x = 0; x < ffsize; x++){
     tmp = fct[x]* fct[x];
     sum += tmp * tmp;
}
free( fct );
return sum;
}

double   alpha( boole f )
{
long long int sum = 0, tmp;
int  * fct = calloc( ffsize, sizeof(int) );
galois x;
for( x = 0; x < ffsize; x++ )
	fct [ x ] = f[x] ? -1 : + 1;

Fourier( fct, ffsize);
for( x = 0; x < ffsize; x++){
     tmp = fct[x]* fct[x];
     sum += tmp * tmp;
}
free( fct );
return (double) sum  / ffsize / ffsize / ffsize;
}

boole getaglboole( boole f, agl s )
{ boole g = getboole( );
  shortvec x;
  for( x = 0; x < ffsize; x++)
          g[x] = f[ aglImage(x, s) ];
  return g;
}

void addboole( boole g, boole f )
{ shortvec x;
        for( x = 0; x < ffsize; x++ )
                g[x] ^= f[x] ;
}
void aglboole( boole g , agl R )
{ boole t =  getaglboole( g,  R );
  shortvec x;
  for( x = 0; x < ffsize; x++ )
          g[x] = t[x];
  free(t);
}




void pANF( FILE*dst, boole f )
// prints the anf of f
{
int x,y,i;
boole aux = getboolecpy ( f );
fprintf(dst, "[ ");
int p = 1;
xform( aux, ffsize);
if ( aux[0] )
        fprintf( dst, "1+" );
for( x = 1; x < ffsize ; x++)
   if ( aux[x] == 1 ) {
        i=0;
        y=x;
        if (p==0) fprintf(dst, "+"); else p=0;
        while (y>0)
              {
              if ( y & 1 ) fprintf(dst, "%c", i + 'a');
              i++;
              y=y>>1;
              }
        }
fprintf(dst, " ]");
free(aux);
}

void part( boole f )
{ int t = degree(f);
  shortvec x;
  puts( "\npart:" );
  uchar p[ffsize];
  while ( t >= 1 ) {
         for( x = 0; x < ffsize; x++ )
                 p[x] = f[x];
         projboole(t, t, p );
         pANF(stdout, p );
         t--;
  }
}

uint64_t  booleto64( boole f )
{ uint64_t res = 0, bit = 1;
  shortvec x;
  for( x = 0; x < ffsize; x++ )
          if ( f[x] ) res ^= bit << x;
  return res;
}


boole  wordtoboole( uint64_t w  )
{  shortvec x;
   boole f = getboole();
   for( x = 0; x < ffsize; x++, w>>=1 )
          f[x] = w & 1;
  return f;
}



boole  filebooleparameters( char *fn , int *cpt )
{
  boole g = NULL;
  char line[1024];
  FILE * src = fopen( fn, "r" );
  if ( ! src ) {
          perror( "loadparm");
          exit(1);
  }
  while (! feof (src) )
    {
      line[0] = '\0';
      fgets (line, 1024, src);
      switch (*line)
        {
        case '#':
          break;
        case 'a' :
          if ( *cpt == 0 ) {
                g = strtoboole ( line );
          }
          *cpt = *cpt + 1;
          break;
        default:;
        }
    }
    fclose( src );
    return g;
}

boole loadaglboolesize(FILE *src, aglGroup *g, uint64_t *size)
//-get the current object function
{
  boole f;
  char line[1024];
  *g = NULL;
  while (! feof (src) )
    {
      line[0] = '\0';
      fgets (line, 1024, src);
      switch (*line)
        {
        case '#':
          break;
        case '[' :
          addgenstr( line, g);
          break;
        case 'a' :
          f = strtoboole ( line );
          break;
        case 's' :
          sscanf( line, "stabSize=%ld", size );
          return f;
          break;
        default:;
        }
    }
    return NULL;
}




boole loadaglboolesizebyidt(FILE *src, aglGroup *g, uint64_t *size, int idt)
//-get the current object function
{
  int num = 0;
  boole f;
  char line[1024];
  *g = NULL;
  while (! feof (src) )
    {
      line[0] = '\0';
      fgets (line, 1024, src);
      switch (*line)
        {
        case '#':
          break;
        case '[' :
          addgenstr( line, g);
          break;
        case 'a' :
          f = strtoboole ( line );
          break;
        case 's' :
          sscanf( line, "stabSize=%ld", size );
          if ( num == idt )
		  return f;
	  free( f );
	  aglfreeGroup( * g );
	  *g = NULL;
	  num++;
          break;
        default:;
        }
    }
    return NULL;
}









