
#include <stdlib.h>
#include <stdio.h>
#include "boolean.h"

#include "code.h"
#include <stdint.h>

uint64_t  binomial( int k, int m)
{
  if ( ( k == 0 ) ||( k == m ) ) return 1;
  if ( ( k > m  ) || ( k < 0 ) ) return 0;
  return binomial(k, m - 1) + binomial( k - 1, m - 1);
}


int  rmdimen( int s, int t, int m)
{ int res = 0;
  int i;
  for( i = s; i <= t; i++ )
        res += binomial( i , m );
  return res;
}

code getcode( int k, int n )
{
int i;
code res;
res.fct  = calloc( k, sizeof(boole) );
res.nbl  = k;
res.nbc  = n;
res.piv  = calloc( k, sizeof(int) );
for( i=0; i < k; i++ ){
	res.fct[i] = calloc( n, sizeof(uchar) );
	res.piv[i] = -1;
}
return res;
}

code copyCode( code cc )
{
int i, j;
code res = getcode( cc.nbl, cc.nbc );
for( i=0; i < cc.nbl; i++ )
   for( j=0; j < cc.nbc; j++ )
	res.fct[i][j] = cc.fct[i][j];
return res;
}

code multicode( code cc, boole f )
{ code res = getcode( cc.nbl, cc.nbc );
  int i, j;
  for( i=0; i < cc.nbl; i++ )
	  for( j = 0; j < cc.nbc; j++ )
		  res.fct[i][j] =  cc.fct[i][j] & f[j];
 return res; 
}
void freecode( code G )
{
int i;
for( i = 0; i < G.nbl; i++ )
	free( G.fct[i] );
free( G.fct );
free( G.piv );
}


void monotoboole( boole f, int u , int q )
{ int x;
  for( x = 0; x < q; x++ )
        f[x] = 0;
  f[u] = 1;
  xform( f, q);
}



code rmcode( int s, int t, int m)
{
code G;
int u,  k = 0;
int dim = rmdimen(s, t, m );
int  q = 1 << m;
G = getcode( dim , q );

for( u= 0; u < q ; u++ )
   if ( weight(u) >= s && weight(u) <=t ) {
     monotoboole( G.fct[k], u , q );
     k++;
}
return G;
}

int pivotage( code G )
{
int r  =  G.nbl;
galois x, p;
boole tmp;
int i  =  0, j;
while ( i < r ){
   for( p = 0; p < G.nbc && G.fct[i][p]==0; p++ );
   if ( p < G.nbc ) {
       G.piv[i] = p;
       for( j = i+1; j < r; j++ )
	  if ( G.fct[j][p] )
	    for( x = 0; x < G.nbc ; x++ )
                G.fct[j][x] ^=  G.fct[i][x];
       i++;
   } else {
	r--;
        tmp = G.fct[i];
        G.fct[i] = G.fct[r];
        G.fct[r] = tmp;
   }
}
return r;
}


code solve(  uchar X[], code G, uchar Y[], int *solvable )
// return the kernel of x -> x.G
// pass  a solution of  x -> x.G = Y 
// KILL  G and Y
{
boole tmp;

int i,  j, p, x;
 
code B  = getcode( G.nbl  , G.nbl );

int r = B.nbl;

for( i = 0; i < B.nbl ; i++ )
	B.fct[i][i] = 1;

i = 0;

while ( i < r ){
   for( p = 0; p < G.nbc && G.fct[i][p]==0; p++ );
   if ( p < G.nbc ) {
       G.piv[ i ] = p;
       for( j = i + 1; j < r ; j++ )
	  if ( G.fct[ j ][ p ] ) {
	    for( x = 0; x < B.nbc ; x++ )
                B.fct[ j ][x] ^=  B.fct[ i ][x];
	    for( x = 0; x < G.nbc ; x++ )
                G.fct[ j ][x] ^=  G.fct[ i ][x];
	  }
       i++;
   } else {
	r--;
        tmp = G.fct[i]; G.fct[i] = G.fct[r]; G.fct[r] = tmp;
        tmp = B.fct[i]; B.fct[i] = B.fct[r]; B.fct[r] = tmp;
   }
}


code K = getcode( B.nbl - r , B.nbc );

for( i = 0; i < K.nbl ; i++ )
     for( j = 0; j < B.nbc; j++ )
   	K.fct[ i ][ j ] = B.fct[ i + r ][ j ];

for( i = 0; i < K.nbc ; i++ )
	X[ i ] = 0;

for( i = 0; i <  r ; i++ ) 
	if (  Y[ G.piv[ i ] ]  ){ 
		for( j = 0; j < K.nbc; j++ ) 
			X[ j ] ^= B.fct[ i ][ j ];
		for( j = 0; j < G.nbc; j++ )
			Y[ j ] ^= G.fct[ i ][ j ];
		}
r = 0;
for( j = 0; j < G.nbc; j++ )
	r +=   Y[ j ] ;
*solvable = (r == 0);
freecode( B );
return K;
}

int fullpivotage( code G )
{
int r  =  G.nbl;
galois x, p;
boole tmp;
int i  =  0, j;
while ( i < r ){
   for( p = 0; p < G.nbc && G.fct[i][p]==0; p++ );
   if ( p < G.nbc ) {
       G.piv[i] = p;
       for( j = 0 ; j < r; j++ )
	  if ( j != i && G.fct[j][p] )
	    for( x = 0; x < G.nbc ; x++ )
                G.fct[j][x] ^=  G.fct[i][x];
       i++;
   } else {
	r--;
        tmp = G.fct[i];
        G.fct[i] = G.fct[r];
        G.fct[r] = tmp;
   }
}
return r;
}
void randPivotage( code G )
	// assuming full rank matrix...
{
int i, j;
galois x, p;
for( i = 0; i < G.nbl; i++ ) {
   do  p = random( ) % G.nbc; while ( G.fct[i][p]==0 ) ;
   G.piv[i] = p;
   for( j = 0; j < G.nbl; j++ )
      if ( i!= j && G.fct[j][p] )
	    for( x = 0; x < G.nbc ; x++ )
                G.fct[j][x] ^=  G.fct[i][x];
   }
}


void addline( code G, int d, int s )
{ int x;
  for( x = 0; x < G.nbc ; x++ )
                G.fct[d][x] ^=  G.fct[s][x];
}	

void perline( code G, int i, int r )
{ unsigned char *tmp;
        tmp = G.fct[i];
        G.fct[i] = G.fct[r];
        G.fct[r] = tmp;
}

code  kernel( code B, code G )
{
int r  =  G.nbl;
code R;
galois p;
int i  =  0, j;
while ( i < r ){
   for( p = 0; p < G.nbc && G.fct[i][p]==0; p++ );
   if ( p < G.nbc ) {
       G.piv[i] = p;
       for( j = i+1; j < r; j++ )
	  if ( G.fct[j][p] ){
	    addline(G, j, i );
	    addline(B, j, i );
	  }
       i++;
   } else {
	r--;
	perline( G, i, r );
	perline( B, i, r );
   }
}
R = getcode( B.nbl - r, B.nbc );
for( i = 0; i < R.nbl; i++ ){
	for( j = 0; j < B.nbc; j++ )
		R.fct[i][j] = B.fct[i+r][j];
}
return R;
}

void pCode( char*msg, code G, code H )
{
int i;
galois x;
printf("\ncode:%dx%d from %s\n", G.nbl, G.nbc, msg ); 
for( i=0; i < G.nbl; i++ ){
   printf("%2d :", G.piv[i]);
   for( x = 0; x < G.nbc; x++ )
	printf(" %d", G.fct[i][x] );
   printf(" -> %2d :", H.piv[i]);
   for( x = 0; x < H.nbc; x++ )
	printf(" %d", H.fct[i][x] );
   printf("\n");
}
}

void codeImage( uchar Y[], uchar X[], code G ){
int i, j;
for( j = 0; j < G.nbc; j++ )
	Y[ j ] = 0;
for( i = 0; i < G.nbl; i++ )
	if ( X[ i ] )	
		for( j = 0; j < G.nbc; j++ )
			Y[ j ] ^= G.fct[i][ j ];
}	

void pline( char*msg, uchar L[], int n )
{
galois x;
printf("\nline:%d from %s\n", n , msg );
for( x = 0; x < n; x++ )
	printf(" %d", L[x] );
   printf("\n");
}


void pcode( char*msg, code G )
{
int i;
galois x;
printf("\ncode:%dx%d from %s\n", G.nbl, G.nbc, msg ); 
for( i=0; i < G.nbl; i++ ){
   printf("%2d:", G.piv[i]);
   for( x = 0; x < G.nbc; x++ )
	printf("%d", G.fct[i][x] );
   printf("\n");
}
}

int reduce( boole vec, code G )
{
int i, p;
galois x;
for( i = 0; i < G.nbl && G.piv[i]>=0 ; i++ ){
   p = G.piv[i];
   if ( vec[p] )
   	for( x = 0; x < G.nbc; x++ )
		vec[x] ^= G.fct[i][x];
}
for( x = 0; x < G.nbc && vec[x]==0; x++ ) ;
return x == G.nbc;
}



void quotient( code W, code R )
{ int i;
	pivotage( R );
	for( i=0; i < W.nbl ; i++ )
		reduce( W.fct[i], R );
	pivotage( W );
}


code loadcode( char *fn )
{
code res;
FILE *src;
char ligne[1024], *ptr;
galois x;
int k, n;
src = fopen( fn, "r" );
assert( src != NULL );
while (fgets(ligne, 1024, src)) {
    switch (ligne[0]) {
      case '#': break;
      case 'c': sscanf(ligne, "code:%dx%d", &k, &n);
                res = getcode( k, n ); 
		k   = 0;
                break;
      case '[': x = 0;
                ptr = ligne;
                while (*ptr) {
                    if (*ptr == '1') res.fct[k][x++] = 1;
                    if (*ptr == '0') res.fct[k][x++] = 0;
		    ptr++;
                }
                k++;
     }
}
fclose(src);
return res;
}

/*
int reduce( code G )
{ int i, j, p, res = ffsize;
  int t;
  for( i = 0; i < G.nbl  ; i++ ) {
	  p = i;
	  while ( p < ffsize && G.fct[i][p] == 0 ) p++;
	  if ( p == ffsize ) return 0;
	  for( j = 0; j < G.nbl; j++ ) {
		  t = G.fct[j][i];
		  G.fct[j][i] = G.fct[j][p];
		  G.fct[j][p] = t;
	  }
	  for( j = i+1; j < G.nbl; j++ ) {
		  if ( G.fct[j][i] && j != i ) 
			  for( p = 0 ; p< ffsize; p++ )
				 G.fct[j][p] ^= G.fct[i][p]; 
	  }
  }

  i = G.nbl - 1; 
  t = 1;
  for( j = G.nbl ; j < ffsize;  j++ )
	    t += G.fct[i][j] ;
    if ( t < res  )  res = t;
  return res;
}
int * normalize( code G )
{
int *pi = calloc( ffsize, sizeof(int) );
galois x;
int i, j, tmp;
for( i = 0; i < G.nbl; i++ ){
   for( x = 0; x < ffsize && G.fct[i][x]==0; x++ );
   assert( x < ffsize );
   pi[ i ] = x;
   for( j =0; j < G.nbl; j++ ){
      tmp = G.fct[j][x];
      G.fct[j][x] = G.fct[j][i];
      G.fct[j][i] = tmp;
   } 
   for( j =0; j < G.nbl; j++ )
	if ( j != i && G.fct[j][i] )
	   for( x = 0; x < ffsize; x++ )
                G.fct[j][x] ^=  G.fct[i][x];
   }
  return pi;
}

code orthogonal( code G )
{
code H;
int *pi;
int i;
galois x;
pi = normalize( G );
//pcode(G);
H = getcode( ffsize - G.nbl );
for( i = 0; i < G.nbl; i++ )
    for( x = G.nbl; x < ffsize; x++ )
	H.fct[ x - G.nbl ][i] = G.fct[i][x];

for( i = 0; i < H.nbl; i++ )
	H.fct[i][i+G.nbl] = 1;
picode(H, pi, G.nbl );
free( pi );
return( H );
} 
int poids(uint64_t x)
{   int wt=0;
    do  {
      if (x & 1)  wt++;
      x = x >> 1;
    } while (x != 0);
    return(wt);
}

code rmtranslate( int s, int t, boole f  )
{
code G;
int u,  k = 0;
int dim = rmdimen(s, t, ffdimen  );

G = getcode( dim + 1  );

for( u=0; u < ffsize; u++ )
   if ( weight(u) >= s && weight(u) <=t ) {
     monotoboole( G.fct[k], u  );
     k++;
}
for( u = 0; u < ffsize; u++ )
    G.fct[ dim ][u] = f[u];
return G;
}

code rmcode( int s, int t  )
{
code G;
int u,  k = 0;
int dim = rmdimen(s, t, ffdimen  );

G = getcode( dim );

for( u=0; u < ffsize; u++ )
   if ( weight(u) >= s && weight(u) <=t ) {
     monotoboole( G.fct[k], u  );
     k++;
}
return G;
}

code catcode( code A, code B )
{ code R;
  int i;
  galois x;
  R = getcode( A.nbl+B.nbl );
  for( i=0; i < A.nbl; i++)
    for( x = 0; x < ffsize; x++)
	R.fct[i][x] = A.fct[i][x];
  for( i=0; i < B.nbl; i++)
    for( x = 0; x < ffsize; x++)
	R.fct[i+A.nbl][x] = B.fct[i][x];
  return R;
}


void picode( code G, int pi[], int n )
{
int i, j, tmp;
for( i = n-1; i>=0; i-- )
   for( j = 0; j < G.nbl; j++ ){
      tmp = G.fct[j][ i ]; G.fct[j][ i ] = G.fct[j][ pi[i] ]; G.fct[j][ pi[i] ] = tmp;
   } 
}

int reduce( code G )
{ int i, j, p, res = ffsize;
  int t;
  for( i = 0; i < G.nbl  ; i++ ) {
	  p = i;
	  while ( p < ffsize && G.fct[i][p] == 0 ) p++;
	  if ( p == ffsize ) return 0;
	  for( j = 0; j < G.nbl; j++ ) {
		  t = G.fct[j][i];
		  G.fct[j][i] = G.fct[j][p];
		  G.fct[j][p] = t;
	  }
	  for( j = i+1; j < G.nbl; j++ ) {
		  if ( G.fct[j][i] && j != i ) 
			  for( p = 0 ; p< ffsize; p++ )
				 G.fct[j][p] ^= G.fct[i][p]; 
	  }
  }

  i = G.nbl - 1; 
  t = 1;
  for( j = G.nbl ; j < ffsize;  j++ )
	    t += G.fct[i][j] ;
    if ( t < res  )  res = t;
  return res;
}
int * normalize( code G )
{
int *pi = calloc( ffsize, sizeof(int) );
galois x;
int i, j, tmp;
for( i = 0; i < G.nbl; i++ ){
   for( x = 0; x < ffsize && G.fct[i][x]==0; x++ );
   assert( x < ffsize );
   pi[ i ] = x;
   for( j =0; j < G.nbl; j++ ){
      tmp = G.fct[j][x];
      G.fct[j][x] = G.fct[j][i];
      G.fct[j][i] = tmp;
   } 
   for( j =0; j < G.nbl; j++ )
	if ( j != i && G.fct[j][i] )
	   for( x = 0; x < ffsize; x++ )
                G.fct[j][x] ^=  G.fct[i][x];
   }
  return pi;
}

code orthogonal( code G )
{
code H;
int *pi;
int i;
galois x;
pi = normalize( G );
//pcode(G);
H = getcode( ffsize - G.nbl );
for( i = 0; i < G.nbl; i++ )
    for( x = G.nbl; x < ffsize; x++ )
	H.fct[ x - G.nbl ][i] = G.fct[i][x];

for( i = 0; i < H.nbl; i++ )
	H.fct[i][i+G.nbl] = 1;
picode(H, pi, G.nbl );
free( pi );
return( H );
} 

code loadcode( char *fn )
{
code res;
FILE *src;
char ligne[1024], *ptr;
galois x;
int k;
src = fopen( fn, "r" );
assert( src != NULL );
while (fgets(ligne, 1024, src)) {
    switch (ligne[0]) {
      case '#': break;
      case 'd': sscanf(ligne, "dim=%d", &k);
                res = getcode( k ); 
		k   = 0;
                break;
      case '[': x = 0;
                ptr = ligne;
                while (*ptr) {
                    if (*ptr == '1') res.fct[k][x++] = 1;
                    if (*ptr == '0') res.fct[k][x++] = 0;
		    ptr++;
                }
                k++;
     }
}
fclose(src);
return res;
}
*/
