#include "invariantools.h"
#include "code.h"
#include "distrib.h"


#define NBC568 20748
#define NBC457 179
#define NBC346 35


inv_t *getInvariant( int d, int r )
{ inv_t *res = malloc( sizeof(inv_t) );
  res->der = d;
  res->res = r;
  res->val = -1;
  return res;
}

static int compare( const void *a,  const void *b)
{
    inv_t* fa, *fb;

  fa = (inv_t*) a;
  fb = (inv_t*) b;
    
   if ( fa->der !=  fb->der )
	 return fa->der  - fb->der;

   if ( fa->res !=  fb->res )
	 return fa->res  - fb->res;

    return 0;
}

/*
static void shownode(const void *nodep, VISIT type, int depth)
{
    invptr datap;
    int v, i, total = 0;

    switch (type) {
    case postorder:
	break;
    case endorder:
	break;
    case preorder:
	printf("\npreorder\n");
	datap = *(invariant **) nodep;
	for (i = 0; i < datap->len; i++) {
	    v = datap->tab[i];
	    printf("%d:", v);
	    total += v;
	}
	printf("\nval=%d total=%d", datap->val, total);
	printf("\n");
	break;
    case leaf:
	printf("\nLeaf\n");
	datap = *(invariant **) nodep;
	for (i = 0; i < datap->len; i++) {
	    v = datap->tab[i];
	    printf("%d:", v);
	    total += v;
	}
	printf("\nval=%d total=%d", datap->val, total);
	printf("\n");
	break;
	break;
    }
}
*/

int* devRestriction( boole f , int q, basis_t *b )
{
shortvec u, p, msk, x, y, w;
int t1, t2, tmp;
int *t = calloc( q, sizeof(int ) );
vector v;
boole F, G;
F = calloc( b->q, sizeof(uchar) );
G = calloc( b->q, sizeof(uchar) );
for( u = 1; u < q; u++ ){
        p = 1;
        while ((p & u) == 0)
                p <<= 1;
        msk =  p  ^  ( q/2 );
        for( x = 0 ; x < q/2; x++ ){
                y = x;
                if ( y & p ) y^=msk;
                w = weight( y & u ) & 1;
                if ( w ) y^= p;
                F[x] = f[y];
		G[x] = f[y^p];
        }
        v = booleVector( F , b);
        t1 = b->table[ v ];
        v = booleVector( G , b);
        t2 = b->table[ v ];
        tmp = t1 < t2 ? t1*10 + t2: t2*10 + t1;
        t[u] = tmp;
}
free(F);
free(G);
return t;
}

int invRestriction( boole f , int q, basis_t *b, void **root, int * count )
{
int tmp;
int *t = calloc( q, sizeof(int ) );
assert( q == b->q * 2);
t = devRestriction( f, q, b );
tmp  = findtable( t, q, root, count, 1);
return tmp;
}

int* devDerivation( boole f , int q, basis_t *b )
{
shortvec u, p, msk, x, y;
int tmp;
int *t = calloc( q, sizeof(int ) );
vector v;
boole der;

der = calloc( q / 2  , sizeof(uchar) );

for( u = 1; u < q; u++ ){
        p = 1;
        while ((p & u) == 0)
                p <<= 1;
        msk =  p ^ ( q / 2) ;
        for( x = 0 ; x < q/2 ; x++ ){
                y = x;
                if ( y & p ) y^=msk;
                der[ x ] = f[ y ] ^ f[y ^ u ];
        }
        v = booleVector( der , b);
        tmp = b->table[ v ];
        t[u] = tmp;
}
free(der);
return t;
}


int invSimpleDerivation( boole f , int q, basis_t *b,  void **r1, int *c1)
{
int tmp;
int* t;
assert( q == b->q * 2 );
t = devDerivation( f, q, b );
tmp = findtable( t, q, r1, c1 , 1);
return tmp;
}

int invSimpleKernelDerivation( boole f , int q, basis_t *b,  void **r1, int *c1)
{
int tmp;
int* t;
assert( q == b->q * 2 );
t = devDerivation( f, q, b );
tmp = findtable( t, q, r1, c1 , 1);
return tmp;
}


int invDerRes( boole f , int q, basis_t *bd, basis_t *br, void **r1, int *c1)
{
int tmp;
int* td, *tr;
assert( q == bd->q * 2 );
assert( q == br->q * 2 );
td = devDerivation( f, q, bd );
tr = devRestriction( f, q, br );
Fourier( tr, q );
glue( tr, td, q);
free(td);
tmp = findtable( tr, q, r1, c1 , 0);
return tmp;
}

int invDoubleDerivationKernel( boole f , int q, basis_t *b,  void **r1, int *c1, void **r2, int * c2, basis_t *bq, void **rk, int * ck )
{
shortvec u, p, msk, x, y;
int tmp, ker;
int *t = calloc( q, sizeof(int ) );
boole der;
assert( q == b->q * 4 );

der = calloc( q / 2  , sizeof(uchar) );

for( u = 1; u < q; u++ ){
        p = 1;
        while ((p & u) == 0)
                p <<= 1;
        msk =  p ^ ( q / 2) ;
        for( x = 0 ; x < q/2 ; x++ ){
                y = x;
                if ( y & p ) y^=msk;
                der[ x ] = f[ y ] ^ f[y ^ u ];
        }
        tmp = invSimpleDerivation( der, q/2, b, r2, c2 );
	ker = invKernel( der , 2, 4, 5, bq  );
	tmp = InvariantValue( tmp, ker, rk, ck  );
        t[u] = tmp;
}
tmp = findtable( t, q, r1, c1 , 1);
free(der);
return tmp;
}

int invDoubleDerRes( boole f , int q, basis_t *bd, basis_t *br,  void **r1, int *c1, void **r2, int * c2 )
{
shortvec u, p, msk, x, y;
int tmp;
int *t  = calloc( q, sizeof(int ) );
int *tr = calloc( q, sizeof(int ) );
boole der;
assert( q == bd->q * 4 );

der = calloc( q / 2  , sizeof(uchar) );

for( u = 1; u < q; u++ ){
        p = 1;
        while ((p & u) == 0)
                p <<= 1;
        msk =  p ^ ( q / 2) ;
        for( x = 0 ; x < q/2 ; x++ ){
                y = x;
                if ( y & p ) y^=msk;
                der[ x ] = f[ y ] ^ f[y ^ u ];
        }
        tmp = invSimpleDerivation( der, q/2, bd, r2, c2 );
        t[u] = tmp;
}

tr = devRestriction( f, q, br );
Fourier( tr, q );
glue( tr, t , q);
free(t);
tmp = findtable( tr, q, r1, c1 , 0);
free(der);
return tmp;
}


int newInvariant( int d, int r, void **rootp, int* count   )
{
inv_t *ptr;
const void *ret;
ptr = getInvariant(  d, r  );

    ret  = tsearch((void *) ptr, rootp , compare );
    if ( ret == NULL)
        exit(EXIT_FAILURE);
    if ((*(inv_t **) ret)->val != ptr->val) {
        free( ptr );
        return 0;
    };
// it's new ! 
    ptr->val = *count;
    *count   = *count + 1;
    printf("\nnew invariant value -> %d \n", ptr->val );
    return 1;
}

int InvariantValue( int d, int r, void **rootp, int* count   )
{
inv_t *ptr;
const void *ret;
ptr = getInvariant(  d, r  );

    ret  = tsearch((void *) ptr, rootp , compare );
    if ( ret == NULL)
        exit(EXIT_FAILURE);
    if ((*(inv_t **) ret)->val != ptr->val) {
        free( ptr );
        return (*(inv_t **) ret)->val;
    };
// it's new ! 
    ptr->val = *count;
    *count   = *count + 1;
    return ptr->val ;
}



int invKernel( boole f, int r, int s, int t, basis_t * bq  )
{
int  cpt, limite;
boole rop = getboole();
int count[20]={0};
int  k, x;
int tmp;
vector v;
code cc = rmcode( 1 , r, bq->m);
code prd = multicode( cc, f  );
for( k=0; k < prd.nbl; k++ )
        rsreduction( prd.fct[k], r+s  , r+t, bq->q);
code ker = kernel( cc, prd );
freecode( cc );
freecode( prd );

limite = 1 << ker.nbl;

for( cpt = 1; cpt < limite; cpt++ ) {
        for( k = 0; ((1<<k) & cpt) == 0 ; k++ );//gray
        for( x = 0; x < ffsize; x++ )
                rop[x] ^=  ker.fct[k][x];
        v = booleVector( rop, bq );
        tmp = bq->table[ v ] ;
        count[ tmp ]++;
}

/*printf("ker = %d -> ", ker.nbl );
for( k=0; k < 5  ; k++ )
   printf(" %d [%d]",  k, count[k]  );
 */
freecode( ker );
free(rop);
return count[1];
}


