#include "agltools.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "boolean.h"
#include "mapping.h"
#include "basistools.h"
#include "orbitools.h"
#include "orbitData.h"



orbitData initData( boole f, int s, aglGroup G, size_t grpSize )
{
orbitData res;
	res.f = f;
	res.t = degree( f );
	res.s = s;

size_t  stabSize, total = 1;

aglGroup stab;
int r = res.t;

	res.W[ r ].grp  = G;
	res.W[ r ].grpSize  = grpSize; 

vector vec;

while ( r >= res.s) {
         //printf("\n#######level:%d\n", r );
	 res.W[r].base   = monomialBasis(r, r,  ffdimen);
	 if ( r == res.t ) {
		vec  = booleVector( f, & res.W[r].base );
         	res.W[r].ldg = aglVectorGroupAction( res.W[r].grp , & res.W[r].base );
	 } else { 
           vec = 0;
           res.W[r].ldg = aglBoundaryGroupAction( f , res.W[r].grp , & res.W[r].base );
	 }

	  res.W[r].route = transversale( vec ,  res.W[r].ldg, & res.W[r].base  );

	  total *= res.W[r].base.orbSize;
	  
	  stabSize = res.W[r].grpSize / res.W[r].base.orbSize ;
          //printf("\norbsize=%ld stabSize=%ld total=%ld\n", res.W[r].base.orbSize, stabSize, total );
         
	  stab = routeStabilizer( res.W[r].route , res.W[r].ldg,  res.W[r].base.mat.nbl , stabSize);
	  
	  
	  res.W[r-1].grp   = stab;
	  res.W[r-1].grpSize = stabSize;
	  r--;
  }
res.orbSize = total;
//printf("\norbitSize:%ld\n", total );
return res;
}

void freeData( orbitData * res )
{ int r;
  vector v;
	for( r = res->s; r <= res->t; r++ ){
		for( v = 0; v < res->W[r].base.size; v++ )
			if ( res->W[r].route[v]  ) free( res->W[r].route[v] ) ;
		free( res->W[r].base.table );
        	free( res->W[r].route );
		aglfreeGroup( res->W[r].grp );
		aglVectorGroupFree( res->W[r].ldg );
	}
}


vector aglVectorImages( vector f, aglVector S, int dim ){
// the image of f by the action S
// to be improved !!!
vector  r=0;
int j;
for( j = 0; j <  dim; j++ )
          if (  BITVECTOR( j ,f) )
                        r^= S[j];
r ^=  S[ dim  ];
return r;

}

void aglRightCompose( agl L, const agl R )
{ agl T = getaglCompose( L ,  R );
  int i;
  for( i = 0; i <= agldimen; i++ )
	  L[i] = T[i];
  free( T );
}

void aglLeftCompose( const agl L, agl R )
{ agl T = getaglCompose( L ,  R );
  int i;
  for( i = 0; i <= agldimen; i++ )
	  R[i] = T[i];
  free( T );
}

agl megacheck( boole h, orbitData *data )
{
if ( data->t != degree( h ) ) 
	return NULL;

int r = data->t;

boole g = getboolecpy( h );

vector v = booleVector( g  , & data->W[ r ].base );
if ( ! data->W[r].base.table[ v  ] ) {
	free(g);
	return NULL;
}
agl R    = aglinverse( data->W[r].route[ v  ]  );
agl A    = R;

aglboole( g , R  );

r = r - 1;
while ( r >= data->s ) {
	boole p = getboolecpy( g );
	addboole( p, data->f );
	projboole( r, r, p );
	v = booleVector( p , & data->W[ r ].base );
	free( p );
	if ( ! data->W[r].base.table[ v  ] ) {
		free(A);
		free(g);
		return NULL;
	}
	R = aglinverse( data->W[r].route[ v  ]  );
	aglRightCompose( A, R );
	aglboole( g , R  );
	free(R);
	r--;
}
free(g);
return A;
}



int ismember( boole h, orbitData *data )
{
if ( data->t != degree( h ) )
        return 0;

int r = data->t;

boole g = getboolecpy( h );

vector v = booleVector( g  , & data->W[ r ].base );
if ( ! data->W[r].base.table[ v  ] ) {
	free(g);
	return 0;
}
agl R    = aglinverse( data->W[r].route[ v  ]  );
agl A    = R;

aglboole( g , R  );

r = r - 1;
while ( r >= data->s ) {
        boole p = getboolecpy( g );
        addboole( p, data->f );
        projboole( r, r, p );
        v = booleVector( p , & data->W[ r ].base );
        free( p );
        if ( ! data->W[r].base.table[ v  ] ) {
                free(A);
                free(g);
                return 0;
        }
        R = aglinverse( data->W[r].route[ v  ]  );
        aglRightCompose( A, R );
        aglboole( g , R  );
        free(R);
        r--;
}
free(g);
free(A);
return 1;
}















