#define _GNU_SOURCE     /* Expose la déclaration de tdestroy() */
#include <search.h>

#include "code.h"
#include "orbit.h"

#include "basistools.h"

#include "orbitools.h"

#include <assert.h>


uchar * orbitable;
agl   * route;

int   orbSize;
int   dimenSpace;
uchar  numero;
size_t spaceSize;



vector booleVector(boole f, basis_t *base)
{
    int64_t res = 0;
    shortvec a;
    for (a = 0; a < base->q ; a++) {
        if ( f[a] )
		res ^= base->dirac[a];
        }
    return res;
}


aglVector getaglVector(int dim )
{
    return calloc( dim , sizeof(vector));
}

aglVector vectorAction( agl s , basis_t *b)
{
    int j;
    shortvec x;
    aglVector res = getaglVector( b->mat.nbl + 1 );
    boole h = calloc( b->q, sizeof(uchar) );

    res[ b->mat.nbl ] = 0;

    for (j = 0; j < b->mat.nbl; j++) {
	for (x = 0; x < b->q ; x++) 
	    h[x] = b->mat.fct[j][aglImage(x, s)];
	res[j] = booleVector(h, b );
	}
    free(h);
    return res;
}


aglVector boundaryAction(boole f, agl s, basis_t * b)
{
    // assuming  f o s = f + lower degree terms...
    int j;
    shortvec x;
    aglVector res = getaglVector( b->mat.nbl  + 1 );
    boole h = calloc( b->q, sizeof(uchar) );

    for (x = 0; x < b->q; x++) 
        h[x] = f[  aglImage(x, s) ] ^ f[x];


    assert( degree(h) <=   b->t );
    res[ b->mat.nbl ] = booleVector( h , b );    
 
    for (j = 0; j < b->mat.nbl; j++) {
        for (x = 0; x < b->q ; x++)
            h[x] = b->mat.fct[j][aglImage(x, s)];
        res[j] = booleVector(h, b );
        }

    free(h);
    return res;
}


aglVector sideAction(boole f, agl s, basis_t * b)
{
    // assuming  f o s = f + lower degree terms...
    int j;
    shortvec x;
    aglVector res = getaglVector( b->mat.nbl  + 1 );
    boole h = calloc( b->q, sizeof(uchar) );

    for (x = 0; x < b->q; x++)
        h[x] = f[  aglImage(x, s) ] ^ f[x];


    res[ b->mat.nbl ] = booleVector( h , b );

    for (j = 0; j < b->mat.nbl; j++) {
        for (x = 0; x < b->q ; x++)
            h[x] = b->mat.fct[j][aglImage(x, s)];
        res[j] = booleVector(h, b );
        }

    free(h);
    return res;
}






void  aglVectorGroupFree( aglVectorGroup l)
{ 
	aglVectorGroup aux;
	while ( l ) {
		aux = l;
		l = l->next;
		free( aux->act);
		free( aux );
	}
}

aglVectorGroup aglBoundaryGroupAction( boole f, aglGroup g, basis_t *b )
{
// builds the generators for the action of
// the class K on the space B(s,t,m)
    aglVectorGroup res = NULL;

    aglGroup aux;


    aglVectorGroup tmp;

    aux = g;
    while (aux != NULL) {
        tmp = malloc(sizeof(enraglVectorGroup)) ;
        tmp->per = aux->per;            // partage de l'espace memoire
        tmp->act = boundaryAction( f, aux->per, b );
        tmp->next = res;
        res = tmp;
        aux = aux->next;
    }
    return res;

}

aglVectorGroup aglVectorGroupAction( aglGroup g, basis_t *b )
{
// builds the generators for the action of
// the class K on the space B(s,t,m)
    aglVectorGroup res = NULL;

    aglGroup aux;
 

    aglVectorGroup tmp;

    aux = g;
    while (aux != NULL) {
	tmp = malloc(sizeof(enraglVectorGroup)) ;	               
	tmp->per = aux->per;		// partage de l'espace memoire
	tmp->act = vectorAction( aux->per, b );		
	tmp->next = res;
	res = tmp;
	aux = aux->next;
    }
    return res;

}


boole vectortoboole( vector w , basis_t *b)
{ boole res = calloc( b->q , sizeof(uchar) );
  shortvec  j, x;
  for( j = 0; j < b->mat.nbl; j++ )
	  if (  BITVECTOR(j,w) )
		 for( x = 0; x < b->q; x++ ) 
		  	res[ x ] ^=  b->mat.fct[j][x];
  return res; 
}

vector aglVectorImage( vector f, aglVector S){
// the image of f by the action S
// to be improved !!!
vector  r=0;
int j;
for( j = 0; j < dimenSpace; j++ )
	  if (  BITVECTOR( j ,f) )
	  		r^= S[j];
r ^=  S[ dimenSpace ];	  		
return r;

}

size_t browse(vector  w, aglVectorGroup lda  )
{
    vector x, y;
    aglVectorGroup aux;
    openVectorStack();
    pushVector( w );
    //orbitable[ w ] = 1; os
    orbitable[ w ] = numero;
    orbSize = 1;
    while ( ptrStack  > 0 ) {
	x = popVector( );
	aux = lda;
	while (aux != NULL) {

	    y = aglVectorImage(x, aux->act);
	 
	    if (orbitable[y] == 0) {
		pushVector( y );
		orbSize++;
		orbitable[y] = numero;
	    }
	    aux = aux->next;
	}
    }
    closeVectorStack();
    return  orbSize;
}

size_t minbrowse(vector  *w, vector * cw, aglVectorGroup lda  )
{
    vector x, y, cx, cy;
    aglVectorGroup aux;
    openVectorStack();
    pushVector( *w  );
    pushVector( *cw );
    orbitable[ *w ] = numero;
    orbSize = 1;
   
    while ( ptrStack  > 0 ) {
	cx = popVector( );
        x  = popVector( );
	if ( x < *w  ) { 
		 *w = x;
		*cw = cx; 
	};
        aux = lda;
        while (aux != NULL) {

            y = aglVectorImage( x, aux->act);
            cy = aglVectorImage(cx, aux->act);

            if (orbitable[y] == 0) {
                pushVector( y );
                pushVector( cy );
                orbSize++;
                orbitable[y] = numero;
            }
            aux = aux->next;
        }
    }
    closeVectorStack();
    return  orbSize;
}


size_t numbrowse(int num, vector  w, aglVectorGroup lda  )
{
	numero = num;
	return browse( w, lda  );
}

int  orbitClassBasic( aglGroup S, basis_t *b, class *r  )
{  aglVectorGroup ldg;
   vector f;
   boole fct;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   orbitable = calloc( spaceSize, sizeof(char));
   ldg =  aglVectorGroupAction( S , b ); 
   numero = 0;
   for( f = 1; f < spaceSize; f++){
                if ( orbitable[f]==0 ) { 
   			 orbSize = 0; 
			 numero++;
                         browse(f, ldg );
			 fct = vectortoboole( f, b );
			 pushClass( fct, NULL, 0 , r );	
                }
                }
   aglVectorGroupFree( ldg );
   b->table = orbitable;
   b->rank  = numero;
   return numero;
}



int orbitBasic(aglGroup S, basis_t *b)
{ aglVectorGroup ldg;
   vector f;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   printf("#action on space of dim : %d\n", dimenSpace );
   orbSize = 0; 
   numero  = 0;
   orbitable = calloc( spaceSize, sizeof(char));
   ldg =  aglVectorGroupAction( S , b ); 
   for( f = 0  ; f < spaceSize; f++){
                if ( orbitable[ f ]==0 ) { 
			 orbSize=0;
			 numero++;
                         browse(f, ldg );
   			 if ( verb ) 
				 printf("\norb %d -> %d", numero, orbSize );
			 assert( numero != 0 );
                }
                }
   aglVectorGroupFree( ldg );
   b->table = orbitable;
   b->rank  = numero;
   return numero;
}


int orbitTrace(aglGroup S, basis_t *b)
{ aglVectorGroup ldg;
   vector f;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   printf("\n#action on space of dim : %d\n", dimenSpace );
   numero  = 2;
   b->rank = 0;
   orbitable = calloc( spaceSize, sizeof(char));
   ldg =  aglVectorGroupAction( S , b );
   for( f = 0 ; f < spaceSize; f++){
                if ( orbitable[f]  == 0 ) {
                         browse(f, ldg );
			 orbitable[ f ] = 1;
			 b->rank++;
                }
                }
   aglVectorGroupFree( ldg );
   b->table = orbitable;
   return numero;
}

void  initBrowse( basis_t *b )
{ 
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   numero  = 2;
   b->rank = 0;
   orbitable = calloc( spaceSize, sizeof(char));
   b->table = orbitable;
}

int nodecmp(const void *pa, const void *pb) {
             if (*(vector *)pa < *(vector *)pb) return -1;
             if (*(vector *)pa > *(vector *)pb) return 1;
             return 0;
           }


size_t orbittree(vector f, aglGroup S, basis_t *b)
{ aglVectorGroup lda, aux;
   vector x, y;
   vector *ptr;
   void * ret;
   void * root = NULL;
   dimenSpace = b->mat.nbl;
   printf("#action on space of dim : %d\n", dimenSpace );
   lda =  aglVectorGroupAction( S , b ); 
   openVectorStack();
   pushVector( f ); 
   ptr  = malloc( sizeof(vector) );
   *ptr = f;
   ret  = tsearch((void *)ptr, &root, nodecmp );
   orbSize = 1;
   while ( ptrStack  > 0 ) {
	x = popVector( );
	aux = lda;
	while (aux != NULL) {

	    y = aglVectorImage(x, aux->act);

	    ptr  = malloc( sizeof(vector) );
	    *ptr = y;
	    ret  = tfind((void *)ptr, &root, nodecmp );
	    if ( ret == NULL ) {
		     ret  = tsearch((void *)ptr, &root, nodecmp );
	             pushVector( y );
		     orbSize++;
	    }
	    aux = aux->next;
	}
	///printf("\norbSize=%d", orbSize );
    }
   
   aglVectorGroupFree( lda );

   closeVectorStack();
   tdestroy( root, free );
   return orbSize;
}




void fullbrowse(vector  w, aglVectorGroup lda  )
{
    vector x, y;
    aglVectorGroup aux;
    openVectorStack();
    pushVector( w );
    
    //orbitable[ w ] = 1; os
    route[ w ] = getaglidt(0) ;
    orbitable[ w ] = numero;
    orbSize = 1;
    while ( ptrStack  > 0 ) {
        x = popVector( );
        aux = lda;
        while (aux != NULL) {

            y = aglVectorImage(x, aux->act);
            if ( route[y] == NULL ) {
                pushVector( y );
		route[y] = getaglCompose( route[x], aux->per ); 
		orbitable[ y ] = numero;
                orbSize++;
            }
            aux = aux->next;
        }
    }
    closeVectorStack();
}


/*
typedef struct {
        form w;
        agl s;
} object;

typedef struct _objectstack_ {
   object info;
   struct _objectstack_ * next;
} enrstack, *stack;


void pushObject( object , stack *);

object  popObject( stack *);
*/


agl getaglorbit(vector w, aglGroup S, basis_t *b, int  limite )
{  aglVectorGroup ldg;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   orbSize = 0;
   numero  = 1;

   orbitable = calloc( spaceSize, sizeof(char));
   ldg =  aglVectorGroupAction( S , b );
    
    aglVectorGroup aux;

    object x, y;
    x.s  = getaglidt(0);
    x.w  = w;
    form  min  = w;
    agl res  = getaglidt( 0 );

    orbitable[ x.w ] = 1;
    orbSize = 1;

    stack pile = NULL;
    pushObject( x, & pile);
    int tour = 0;
    while ( pile  && tour < limite ) {
        x = popObject( &pile );
        aux = ldg;
        while (aux != NULL) {
            y.w = aglVectorImage(x.w , aux->act);
            if ( orbitable[ y.w ] == 0 ) {
		y.s  = getaglCompose( x.s, aux->per ); 
                pushObject( y, &pile  );
		orbitable[ y.w  ] = 1;
		if ( y.w < min ) {
			int i;
			for( i = 0; i <= agldimen; i++ )
			       res[i] = y.s[i];
			min = y.w;
		}
            }
            aux = aux->next;
        }
	free( x.s  );
	tour++;
    }

    while ( pile ) {
	    x = popObject( &pile );
	    free( x.s );;
    }
    //printf("tour=%d\n", tour );
    free( orbitable);
    return res;
}

agl* orbitFull(aglGroup S, basis_t *b)
{ aglVectorGroup ldg;
   vector f;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   printf("#action on space of dim : %d\n", dimenSpace );
   orbSize = 0; 
   numero  = 0;
   route = calloc( spaceSize, sizeof(agl));
   orbitable = calloc( spaceSize, sizeof(char));
   ldg =  aglVectorGroupAction( S , b ); 
   for( f = 0 ; f < spaceSize; f++){
                if ( route[ f ]== NULL ) { 
			 orbSize=0;
			 numero++;
                         fullbrowse(f, ldg );
   			 //printf("\norb %d -> %d", numero, orbSize );
			 assert( numero != 0 );
                }
                }
   aglVectorGroupFree( ldg );
   b->rank  = numero;
   b->table = orbitable;
   return route;
}

              



void repbrowse(vector  w, aglVectorGroup lda, rep_t *rep  )
{
    vector x, y;
    aglVectorGroup aux;
    openVectorStack();
    pushVector( w );
    rep[ w ] = w;
    orbSize = 1;
    while ( ptrStack  > 0 ) {
	x = popVector( );
	aux = lda;
	while (aux != NULL) {

	    y = aglVectorImage(x, aux->act);
	 
	    if ( rep[y] == 0 ) {
		pushVector( y );
		rep[ y ] = w;
	    }
	    aux = aux->next;
	}
    }
    closeVectorStack();
}


rep_t * orbitRepresentative(aglGroup S, basis_t *b)
{ aglVectorGroup ldg;
   vector f;
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   orbSize = 0; 
   numero  = 0;
   rep_t * res  = calloc( spaceSize, sizeof(rep_t));
   ldg =  aglVectorGroupAction( S , b ); 
   for( f = 1 ; f < spaceSize; f++){
                if ( res[ f ] == 0 ) { 
			 orbSize=0;
                         repbrowse(f, ldg , res );
                }
                }
   aglVectorGroupFree( ldg );
   b->rank  = numero;
   b->table = orbitable;
   return res;
}
void controle( vector f,  aglVectorGroup ldg, basis_t *b )
{
	aglVectorGroup tmp = ldg;
	vector y;
	boole g;
	printf("\ndim=%d", dimenSpace );
	while ( tmp ){
                  pagl( stdout, tmp->per );
                  y = aglVectorImage( f , tmp->act);
                  printf("\nx=%ld ---> y=%ld\n", f , y );
                  g = vectortoboole( y , b );
                  panf(stdout,  g);

                  tmp = tmp->next;
          }
}
agl* transversale(vector f, aglVectorGroup ldg, basis_t *b)
{ 
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   //printf("\n#transersale  dimension : %d\n", dimenSpace );
   orbSize = 0;
   numero  = 1;
   route = calloc( spaceSize, sizeof(agl));
   orbitable = calloc( spaceSize, sizeof(char));
   orbSize = 0;

         fullbrowse(f, ldg );

   //printf("#transersale orbitSize: %d\n", orbSize );
   b->rank  = numero;
   b->orbSize  = orbSize;
   b->table = orbitable;
   return route;
}

vector  minfullbrowse(vector  w, aglVectorGroup lda  )
{
    vector x, y;
    aglVectorGroup aux;
    openVectorStack();
    pushVector( w );
    
    //orbitable[ w ] = 1; os
    route[ w ] = getaglidt(0) ;
    orbitable[ w ] = numero;
    orbSize = 1;
    vector r = w;
    while ( ptrStack  > 0 ) {
        x = popVector( );
        aux = lda;
        while (aux != NULL) {

            y = aglVectorImage(x, aux->act);
            if ( route[y] == NULL ) {
                pushVector( y );
		route[y] = getaglCompose( route[x], aux->per ); 
		orbitable[ y ] = numero;
                orbSize++;
		if ( y < r ) r = y;
            }
            aux = aux->next;
        }
    }
    closeVectorStack();
    return r;
}

agl minroute(vector f, aglVectorGroup ldg, basis_t *b)
{
   dimenSpace = b->mat.nbl;
   spaceSize  = ( (size_t) 1 ) <<   dimenSpace;
   b->size  = spaceSize;
   numero  = 1;
   route = calloc( spaceSize, sizeof(agl));
   orbitable = calloc( spaceSize, sizeof(char));
   orbSize = 0;
   vector  r = minfullbrowse( f, ldg );
   agl res  = aglinverse( route[r] );
   vector v;
   for( v = 0; v < spaceSize ; v++ )
	   if ( route[v] ) free( route[v] );
   free(route);
   return res;
}



static int
       compare(const void *pa, const void *pb)
       {
           const object * pA, *pB;
           pA = pa;
           pB = pb;

           if ( pA->w  < pB->w )
               return -1;
           if ( pA->w  > pB->w )
               return 1;
           return 0;
       }


aglGroup plainStabilizer(vector f, aglGroup S, basis_t *b, size_t but)
{  // fing generator set of w stabilizer
    aglGroup     res = NULL;
    stack   p = NULL;
    aglVectorGroup lda, aux;
    
    agl A, I;
 
    int size = 0;
    void *root = NULL;
    void *val;
    object *ptr;
    object x;
    
    dimenSpace = b->mat.nbl;
    
    //printf("#search stab of order %lu on space of dim : %d\n", but, b->mat.nbl  );
    lda =  aglVectorGroupAction( S , b );

    x.w = f ;
    x.s = getaglidt(0);
    pushObject( x,  &p  );
    size = 1;
    ptr = malloc( sizeof( object) );
    ptr->w  = f;
    ptr->s  = getaglidt(0);

    //phil : initialisation...
    val = tsearch( (void*) ptr, &root, compare);

    ssinit();
    while (  size > 0 && ssOrder() != but  ) {
	x = popObject( & p  );
	size--;
	aux = lda;
	while (aux != NULL) {
	    ptr = malloc( sizeof( object) );
	    ptr-> w = aglVectorImage(x.w, aux->act);
	    ptr-> s = getaglCompose(x.s, aux->per);
	    val = tsearch((void *) ptr, &root, compare);
	    if (val == NULL)
		exit(EXIT_FAILURE);

	    if ((*(object **) val) != ptr) {
		// deja vu
            	I = aglinverse( (* (object**) val)->s );
            	A = getaglCompose( ptr->s , I );
                free(I);
                SchreierSims(A, 0);
		//phil : why ??
		free(ptr);
	        free(A);
	    } else {
		// nouveau
		pushObject( *ptr, & p );
		size++;
		//phil : why
		//free(ptr);
	    }

	    aux = aux->next;
	}
    }
    while ( p )
	popObject( &p);

    //tdestroy( &root, freenode );
    while (root != NULL) {
        ptr = *((object **)root);
        tdelete(ptr, &root, compare);
        free(ptr->s);
        free(ptr);
    } 
    res = ssgen();
    ssfree();
    aglVectorGroupFree( lda );
    return res;
}



aglGroup routeStabilizer(agl* R, aglVectorGroup ldg, int dim, size_t but)
{  // fing generator set of w stabilizer
    aglGroup     res = NULL;
    aglVectorGroup aux; 
    agl  Ry, T, I;
 
    size_t size = 1;
    size <<= dim;  
    vector x, y;  
   // printf("\n#route stab of order %lu on space of dim %d...\n", but, dim  );

    ssinit();
 
    x = random() % size; 
    while ( ssOrder() <  but  ) {
	        do 
			x = ( x + 1 ) % size;
		while ( ! R[x] ) ;

                aux = ldg;
	        while (  aux != NULL ) {
            		y = aglVectorImage( x, aux->act);
			Ry = getaglCompose( R[x], aux->per  );
			I  = aglinverse(  Ry  );
			T  = getaglCompose( R[y], I  );
                	SchreierSims( T , 0);
			free(I);
			free(T);
			free(Ry);
			aux = aux->next;
		}
            	
	    }

    assert( but == ssOrder() );
    res = ssgen();
    ssfree();
    return res;
}





