
#include "spacetools.h"
#include "code.h"

#include "orbit.h"
#include <assert.h>



int spaceDimen, spaceValuation;
size_t spaceSize;
int *monomial;
int *number;
int spaceDegree;
space * dirac;



void initSpace(int s, int t, int m)
{
    int j = 0;
    shortvec u;
    spaceDimen = rmdimen(s, t, m);
    spaceSize = 1ULL << spaceDimen;
    spaceDegree = t;
    spaceValuation = s;
    monomial = calloc(spaceDimen, sizeof(int));
    number = calloc(ffsize, sizeof(int));
    for (u = 0; u < ffsize; u++) {
	number[u] = -1;
	if (weight(u) <= spaceDegree && weight(u)>= spaceValuation ) {
	    monomial[j] = u;
	    number[u] = j;
	    j++;
	}
    }
    printf("\n#Working space : %d size %ld\n", spaceDimen, spaceSize );
}

void closeSpace(void)
{
    free(number);
    free(monomial);
}

int64_t oldsmallboole(boole f )
// assume to work  in codim 2
{
    int64_t res = 0;
    shortvec u;
    xform( f, ffsize/4 );
    for (u = 0; u < ffsize/4; u++) {
        if (f[u] && (number[u] >= 0)) {
            res ^= 1LU << number[u];
        }
    }
    xform( f, ffsize/4 );
    return res;
}
int64_t smallboole(boole f )
// assume to work  dimension 8...
{
    int64_t res = 0;
    shortvec a;
    for (a = 0; a < ffsize/4; a++) {
        if ( f[a] )
		res ^= dirac[a];
        }
    return res;
}

int64_t prepareDirac( void  )
// assume to work  in dimension 6 !!
{
assert( ffdimen == 6 );
    int64_t res = 0;
    boole f;
    dirac  = calloc( ffsize  , sizeof(int64_t) );
    shortvec a, u;
    for( a = 0; a < ffsize ; a++ ) {
	    f = getboole( );
	    f[a] = 1;
            xform( f, ffsize  );
	    res = 0;
            for (u = 0; u < ffsize ; u++) {
       		if (f[u] && (number[u] >= 0)) 
                    res ^= 1LU << number[u];
    	   }
	    dirac[ a ] = res;
    		free( f );
    }
    return res;
}



space booletospace(boole f)
{				// return a space from f and transspace f to ANF
    space res = 0;
    shortvec u;
    TTtoANF(f);
    for (u = 0; u < ffsize; u++) {
	if (f[u] && (number[u] >= 0)) {
	    res ^= 1LU << number[u];
	}
    }
    ANFtoTT(f);
    return res;


}


aglSpace getaglSpace(void)
{
    return calloc(spaceDimen, sizeof(space));
}

aglSpace spaceAction( agl s )
{
    int j;
    shortvec x, u;
    aglSpace res = getaglSpace();
    boole g = getboole();
    boole h = getboole();

    for (j = 0; j < spaceDimen; j++) {
	for (u = 0; u < ffsize; u++) {
	    g[u] = 0;
	}
	g[monomial[j]] = 1;
	ANFtoTT(g);
	for (x = 0; x < ffsize; x++) {
	    h[x] = g[aglImage(x, s)];
	}
	res[j] = booletospace(h);
    }
    free(g);
    free(h);
    return res;
}

void  aglSpaceGroupFree( aglSpaceGroup l)
{ 
	aglSpaceGroup aux;
	while ( l ) {
		aux = l;
		l = l->next;
		free( aux->act);
		free( aux );
	}
}

aglSpaceGroup aglSpaceGroupAction( aglGroup g)
{
// builds the generators for the action of
// the class K on the space B(s,t,m)
    aglSpaceGroup res = NULL;

    aglGroup aux;
 

    aglSpaceGroup tmp;

    aux = g;
    while (aux != NULL) {
	tmp = malloc(sizeof(enraglSpaceGroup)) ;	               
	tmp->per = aux->per;		// partage de l'espace memoire
	tmp->act = spaceAction( aux->per );		
	tmp->next = res;
	res = tmp;
	aux = aux->next;
    }
    return res;

}

// return the sum of the Boolean function fct and the space w.
boole spacetoboole( space w )
{ boole res = getboole();
  shortvec  j;
  for( j = 0; j < spaceDimen; j++ )
	  if (  BITSPACE(j,w) ) 
		  res[ monomial[j]  ] =  1;
  ANFtoTT( res );
 
  return res; 
}

space aglSpaceImage( space f, aglSpace S){
// the image of f by the action S
space  r=0;
int j;
for( j = 0; j < spaceDimen; j++ )
	  if (  BITSPACE(j,f) )
	  		r^= S[j];
	  		
return r;

}

char * table;

int count;
/*
static void oldbrowse (space f, aglSpaceGroup listGen)
{
aglSpaceGroup aux;
space  g;
count++;
table[ f ] = 1;
aux = listGen;
while (aux !=NULL)
        {
        g= aglSpaceImage( f, aux->act );
        if (table[g]==0)
                browse(g,listGen);
        aux = aux->next;
        }
return;
}
*/

void browse(space  w, aglSpaceGroup lda, space *trs )
{
    space x, y;
    int i;
    aglSpaceGroup aux;
    openSpaceStack();
    pushSpace( w );
    table[ w ] = 1;
    while ( ptrStack  > 0 ) {
	x = popSpace( );
	aux = lda;
	for( i = 0; i <= ffdimen ; i++ ){
		y  = x ^ trs[i];
		if (table[y] == 0) {
                   pushSpace( y );
                   table[y] = 1;
                }
	}
	while (aux != NULL) {

	    y = aglSpaceImage(x, aux->act);
	    if (table[y] == 0) {
		pushSpace( y );
		table[y] = 1;
	    }
	    aux = aux->next;
	}
    }
    closeSpaceStack();
}


boole xmul( int i, boole g )
{ boole res = getboole();
  shortvec x;
  for ( x = 0; x < ffsize; x++ )
	res[x] = ( x & (1<<i) ) ?  g[x] : 0;
  return res;  
}

space * mktrans( boole g )
{ space *res = calloc( ffdimen+1, sizeof(space) );
  int i;
  boole f;
  for( i = 0; i < ffdimen; i++ ) {
	  f = xmul( i , g );
	  res[i] =  booletospace( f );
	  free(f);
  }
  res[ ffdimen ] = booletospace( g );
return res;
}

int  orbitSpace( aglGroup S, boole g, class *r )
{ aglSpaceGroup ldg;
 table = calloc( spaceSize, sizeof(char));
 space f;
 space *trs;
 boole fct;
 ldg =  aglSpaceGroupAction( S ); 
 trs =  mktrans( g ); 
 int count = 0;
 for(f=0; f < spaceSize; f++){
                if (table[f]==0) { 
                         browse(f, ldg, trs);
			 count++;
			 fct = spacetoboole( f );
			 pushClass( fct, NULL, 0 , r );	
                }
                }
aglSpaceGroupFree( ldg );
free(table);
return count;
}






