#include "agltools.h"
#include "sstools.h"
#include <assert.h>


record *ss = NULL, *save = NULL;

void ssswap( void )
{ record *tmp;
  tmp  = save;
  save = ss;
  ss   = tmp; 
}


void SchreierSimsAction( shortvec x, agl A, char tt[], agl by[], aglGroup g)
{ shortvec y;
  aglGroup t;
  agl B;
  if ( tt[x] ) {
	puts("panic!");
  }
  tt[x] = 1;
  by[x] = getaglcpy( A );
  t = g;
  while ( t ) {
    y = aglImage(x, t->per );
    if ( tt[y] == 0 ){ 
	B  = getaglCompose(  t->per, A );
        SchreierSimsAction( y, B, tt, by, g);
	free(B);
    }
    t = t->next;
  }
}

ullong mkaglorb( shortvec y, char *tt , agl by[], aglGroup g)
{ shortvec x;
  agl A;
  ullong r;
  for( x = 0; x < aglsize; x++) {
     if ( tt[x] ) free( by[x] );
     tt[x] = 0;
  }
  A = getaglidt( 0 );
  SchreierSimsAction( y, A, tt, by, g);
  free(A); 
  r = 0;
  for( x = 0; x < aglsize; x++)
    if ( tt[x] ) r++;
  return r;
}

int strip( agl A , int i )
{ agl tmp, B; 
  shortvec y;
  int j;
  for(       ; i <= agldimen; i++ ){
    y = aglImage( ss[i].bp, A );
    if ( ss[i].tt[y] == 0 ) return i;
    tmp = aglinverse( ss[i].by[y] );
    B   = getaglCompose( tmp, A ); //check and improve
    for( j=0; j <= agldimen; j++ )
		A[j] = B[j];
    free( tmp );
    free( B );    // to be improve 
  }
  return agldimen + 1;
}


void SchreierSims( agl B , int i )
{ int l;
  agl A, T, tmp;
  aglGroup g;
  int x, y;
 
  if ( i > agldimen ) return;

  
  A = getaglcpy( B );
  l = strip(  A , i );
  if ( l <= agldimen  ){
    addgen(A, & ss[i].ldg ); // to check
    ss[i].count = mkaglorb( ss[i].bp, ss[i].tt, 
                  ss[i].by, ss[i].ldg );
    g = ss[i].ldg;
    while ( g ) {
      for( x = 0; x < aglsize; x++)
	if ( ss[i].tt[x] ) {
         y = aglImage( x, g->per);
           
         T = aglinverse( ss[i].by[ y ] );   
	 tmp = getaglCompose(  T, g->per );   
	 free( T );
	 T = getaglCompose( tmp, ss[i].by[ x ] );
	 free( tmp );
         SchreierSims( T , i+1 );
         free(T);
      }
      g = g -> next;
    }
  } else free( A ); 
}

int isingroup( agl B )
{
 int i = 0;
 shortvec y;
 agl A = getaglcpy( B ), tmp; 
  for(       ; i <= agldimen; i++ ){
    y = aglImage( ss[i].bp, A );
    if ( ss[i].tt[y] == 0 ) {
		free(A);
		return 0;
    }
    tmp = aglComposebyinverse( ss[i].by[y], A  );
    free(A);
    A = tmp;
  }
  free(A);
  return 1;

}


void showss( void )
{
int i;
  shortvec x;

  for( i = 0; i <= agldimen; i++){
        printf("\n%d : ",i );
        for(x=0; x < aglsize; x++)
		printf("%2d",  ss[i].tt[x] );
  }

}

void ssinit( void  )
{ int i;
  ss = calloc( agldimen+1, sizeof(record) );

  for( i = 0; i <= agldimen; i++){
	ss[i].tt  = calloc( aglsize, sizeof(char) );
	ss[i].by  = calloc( aglsize, sizeof(agl) );
        if ( i < agldimen ) ss[i].bp = 1 << i;
        else ss[i].bp = 0;
        ss[i].by[ ss[i].bp ] = getaglidt(0); // ???
       ss[i].tt[ ss[i].bp ] = 1; //
       ss[i].count = 1;// 
       ss[i].ldg = NULL;
  }
}

void cleanss( void )
{ int i;
  for( i = 0; i <= agldimen; i++){
    aglfreeGroup( ss[i].ldg);
    ss[i].ldg = NULL;
  }
  ssinit(  );
}

void ssfree( void )
{ int i;
  shortvec x;
  for( i = 0; i <= agldimen; i++){
    aglfreeGroup( ss[i].ldg);
    ss[i].ldg = NULL;
    for( x = 0; x < aglsize; x++ )
	if ( ss[i].tt[x] ) free( ss[i].by[x] );
    free( ss[i].tt );
    free( ss[i].by );
  }
  free( ss );
  ss = NULL;
}

void ssGroup( aglGroup g )
{ ssinit( );
  while ( g ) {
    SchreierSims( g->per , 0 );
    g = g->next;
  }
}
int  nbgenss( void )
{
	aglGroup aux = ss[0].ldg;
	int res = 0;
	while ( aux ) {
		res++;
		aux = aux->next;
	}
	return res;
}
aglGroup  ssgen( void  )
{
  return  aglGroupcpy( ss[0].ldg );
}

agl randss( record * ss)
{
int i;
shortvec x;
agl r = getaglidt(0);
for(i=0; i <= agldimen ; i++ ){
   do {
        x = random() % aglsize;
   } while ( ss[i].tt[x] == 0 );
   r = getaglCompose(  r, ss[i].by[x] );
}
return r;
}

uint64_t  ssOrder( void )
{ int i;
  uint64_t r = 1;
  for( i = 0; i <=agldimen; i++ )
    r *= ss[i].count;
  return r;
}

void  mpz_ssOrder( mpz_t res )
{ int i;
  mpz_set_ui( res, 1 );
  for( i = 0; i <=agldimen; i++ ){
    mpz_mul_ui( res, res, ss[i].count );
  }
}

void  mpz_aglorder( mpz_t res )
{ int i;
  mpz_init( res );
  mpz_set_ui( res, 1 << agldimen );
  for( i = 0; i < agldimen; i++ ){
    mpz_mul_ui( res, res, (1 << agldimen) - (1 << i) );
  }
}

agl * sstable;


uint64_t  ssno;
void fill( agl A, int r  )
      {
	      agl B;
      shortvec x;
      for( x = 0; x <aglsize; x++ )
	      if ( ss[r].tt[x] ) {
		      B = getaglCompose( A,  ss[r].by[x] );
		      if ( r < agldimen ) {
			      fill( B, r+1 );
			      free(B); 
		      }
		      else  sstable[ssno++] =  B;
	      }
      }

void fillstart( void )
      {
      shortvec x;;
      for( x = 0; x < aglsize; x++ )
	      if ( ss[0].tt[x] )
		      fill( ss[0].by[x], 1 ); 
      }

agl* devGroup( aglGroup S, uint64_t * size )
{
ssinit( );
ssGroup( S );
*size = ssOrder();
sstable = calloc( *size, sizeof(agl) );
ssno = 0;
fillstart( );
ssfree();
assert( ssno == *size );
return sstable;
}




